/*
 * Misc - Local helper functions
 *
 * Copyright 2017 IBM Corp.
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef MISC_H
#define MISC_H

void misc_str_remove_symbol(char *str, char symbol);

#endif /* MISC_H */
