/*
 * Copyright IBM Corp. 2001, 2018
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef OUTPUT_H
#define OUTPUT_H

#include <stdio.h>

int write_dump(FILE *stream);

#endif /* OUTPUT_H */
