/*
 * tunedasd - Adjust tunable parameters on DASD devices
 *
 * Copyright IBM Corp. 2004, 2017
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef TUNEDASD_H
#define TUNEDASD_H

#include <getopt.h>
#include <stdint.h>
#include <unistd.h>

void error_print (const char* fmt, ...);

#endif /* not TUNEDASD_H */
