/*
 * zdev - Modify and display the persistent configuration of devices
 *
 * Copyright IBM Corp. 2025
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef VIRTIO_H
#define VIRTIO_H

struct devtype;
struct subtype;

extern struct devtype virtio_devtype;
/* Subtypes are defined in virtio_devtype.subtypes array */

#endif /* VIRTIO_H */
